from dataclasses import dataclass
from typing import Dict, Any
from typing import List, Type, Callable, Generic

from pydantic import BaseModel
from sanic import Sanic, HTTPResponse
from sanic.router import Router

from web_foundation.extentions.addon_loader import AddonsLoader
from web_foundation.extentions.openapi_spec import add_openapi_spec
from web_foundation.extentions.request_handler import Protector, HandlerType, GenericRequestHandler
from web_foundation.kernel.container import GenericDependencyContainer


@dataclass
class RouteMethodConf:
    method_name: str
    protector: Protector | None
    in_dto: Type[BaseModel] | None
    out_dto: Type[BaseModel] | None
    handler: HandlerType
    response_fabric: Callable[[Any], HTTPResponse] | None
    version_prefix: str | None
    version: str | None


@dataclass
class RouteConf:
    app_name: str
    path: str
    methods: List[RouteMethodConf]


class ExtRouter(Router, Generic[GenericRequestHandler]):
    ext_handler: GenericRequestHandler
    parsed_routes: List[RouteConf]

    def __init__(self, ext_handler: GenericRequestHandler):
        super().__init__()
        self.ext_handler = ext_handler
        self.parsed_routes = []

    def apply_routes(self, app: Sanic, container: GenericDependencyContainer, *args, **kwargs):
        raise NotImplementedError


class DictRouter(ExtRouter, Generic[GenericRequestHandler]):
    _router_conf: Dict
    addon_loader: AddonsLoader | None
    versioning: bool
    open_api: bool

    def __init__(self,
                 routes_config: Dict,
                 ext_handler: GenericRequestHandler,
                 versioning: bool = True,
                 open_api: bool = True):
        super().__init__(ext_handler)
        self._router_conf = routes_config
        self.versioning = versioning
        self.open_api = open_api
        self._parse()

    def _parse(self):
        for app_route in self._router_conf.get("apps"):
            version_prefix = app_route.get("version_prefix")
            version_prefix = version_prefix if version_prefix else "/api/v"
            for endpoint, versions in app_route.get("endpoints").items():
                for version, params in versions.items():
                    methods_confs = []
                    endpoint_handler = params.pop("handler", None)
                    endpoint_protector = params.pop("protector", None)
                    endpoint_response_fabric = params.pop("response_fabric", None)
                    for method_name, method_params in params.items():
                        target_func = method_params.get('handler')
                        target_func = target_func if target_func else endpoint_handler
                        protector = method_params.get("protector")
                        protector = protector if protector else endpoint_protector
                        in_dto = method_params.get("in_dto")
                        out_dto = method_params.get("out_dto")
                        response_fabric = method_params.get("response_fabric")
                        response_fabric = response_fabric if response_fabric else endpoint_response_fabric

                        methods_confs.append(RouteMethodConf(method_name=method_name,
                                                             protector=protector,
                                                             in_dto=in_dto,
                                                             out_dto=out_dto,
                                                             handler=target_func,
                                                             response_fabric=response_fabric,
                                                             version_prefix=version_prefix,
                                                             version=version
                                                             ))

                    route = RouteConf(app_name=app_route.get("app_name"),
                                      path=endpoint,
                                      methods=methods_confs)
                    self.parsed_routes.append(route)

    def apply_routes(self, app: Sanic, container: GenericDependencyContainer = None, addon_manager: AddonsLoader = None,
                     **kwargs):
        for route_conf in self.parsed_routes:
            for method_conf in route_conf.methods:
                if method_conf.response_fabric:
                    chain = self.ext_handler(protector=method_conf.protector,
                                             in_struct=method_conf.in_dto,
                                             addon_manager=addon_manager,
                                             response_fabric=method_conf.response_fabric,
                                             container=container)(method_conf.handler)
                else:
                    chain = self.ext_handler(protector=method_conf.protector,
                                             addon_manager=addon_manager,
                                             in_struct=method_conf.in_dto,
                                             container=container)(method_conf.handler)
                if self.open_api:
                    chain = add_openapi_spec(uri=route_conf.path,
                                             method_name=method_conf.method_name,
                                             func=method_conf.handler,
                                             handler=chain,
                                             in_dto=method_conf.in_dto,
                                             out_dto=method_conf.out_dto)

                if self.versioning:
                    app.add_route(uri=route_conf.path,
                                  methods={method_conf.method_name.upper()},
                                  handler=chain,
                                  version_prefix=method_conf.version_prefix,
                                  version=method_conf.version)
                else:
                    app.add_route(uri=route_conf.path,
                                  methods={method_conf.method_name.upper()},
                                  handler=chain)


from web_foundation.errors.app.application import ApplicationError


class SectionNotFound(ApplicationError):
    pass


class FileNotExist(ApplicationError):
    pass


class NothingToWrite(ApplicationError):
    pass


class OsIOError(ApplicationError):
    pass


class NestedFolderDetected(ApplicationError):
    pass



from __future__ import annotations

import io
import json
from dataclasses import dataclass, field
from typing import Any
from typing import Dict

from sanic.response import BaseHTTPResponse

from web_foundation.extentions.realtime.rt_connection import RtMessage, WriteableObj, RtConnection,RtEventCallback
from web_foundation.extentions.request_handler import InputContext


@dataclass
class SseRtMessage(RtMessage):
    _SEPARATOR = "\r\n"
    event_name: str = field(default="")
    data: Dict | None = field(default=None)
    retry: int | None = field(default=None)

    def _prepare(self, *args, **kwargs) -> str | bytes:
        buffer = io.StringIO()
        buffer.write('event: ' + self.event_name + self._SEPARATOR)
        if self.event_id:
            buffer.write('id: ' + str(self.event_id) + self._SEPARATOR)
        if self.retry:
            buffer.write('retry: ' + str(self.retry) + self._SEPARATOR)
        else:
            buffer.write("retry: " + "0" + self._SEPARATOR)
        if self.data:
            buffer.write('data: ' + json.dumps(self.data))
        else:
            buffer.write('data: {}')
        buffer.write("\r\n\r\n")
        return buffer.getvalue()

    @classmethod
    def ping_message(cls):
        return cls(event_id="0", event_name="ping").to_sent


class WriteableSse(WriteableObj):
    obj: BaseHTTPResponse

    def __init__(self, obj: BaseHTTPResponse):
        super().__init__(obj)

    async def write(self, message: Any) -> None:
        await self.obj.send(message)


class SseRtConnection(RtConnection):
    writeable: WriteableSse

    def _ping_msg(self) -> str | bytes:
        return SseRtMessage.ping_message()

    @classmethod
    async def accept_connection(cls, input_ctx: InputContext,
                                resolve_callback: RtEventCallback = None,
                                headers=None,
                                ping_enable: bool = True, ping_timeout=None,
                                listen_timeout=None):
        async with cls(
                input_ctx=input_ctx,
                writeable=await cls._construct_writable(input_ctx, headers=headers),
                resolve_callback=resolve_callback,
                ping_enable=ping_enable,
                ping_timeout=ping_timeout,
                listen_timeout=listen_timeout
        ) as connection:
            await input_ctx.request.app.ctx.rt_broadcaster.accept_rt_connection(
                conn=connection
            )

    @classmethod
    async def _construct_writable(cls, input_ctx: InputContext, headers=None, *args, **kwargs) -> WriteableSse:
        if not headers:
            headers = {}
        headers.update({"X-Accel-Buffering": "no"})
        _response = await input_ctx.request.respond(content_type="text/event-stream; charset=utf-8",
                                                    headers=headers)
        return WriteableSse(_response)

    def _close_condition(self) -> bool:
        if not self.writeable.obj.stream:
            return True
        else:
            return False



from abc import ABCMeta
from enum import Enum
from typing import Any, Dict, TypeVar, Generic, Type

from web_foundation.kernel.resource import Resource


class FilesResource(Resource, metaclass=ABCMeta):
    root: Any

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    async def open(self, path: Any, *args, **kwargs) -> Any:
        raise NotImplementedError

    async def list(self, path: Any, *args, **kwargs):
        raise NotImplementedError

    def exists(self, path: Any) -> bool:
        raise NotImplementedError



from web_foundation.environment.services.service import Service
from web_foundation.environment.workers.web.ext.addons_loader import AddonsLoader


class ApiAddonsService(Service):
    addons_loader: AddonsLoader

    def __init__(self, addons_loader: AddonsLoader):
        self.addons_loader = addons_loader

    async def add_new_middleware(self, filename: str):
        await self.addons_loader.add_new_middleware(filename)

    async def delete_middleware(self, filename: str):
        await self.addons_loader.delete_middleware(filename)



from __future__ import annotations

from typing import TypeVar

from pydantic import BaseSettings

from web_foundation.config import GenericConfig
from web_foundation.kernel.channel import IChannel
from web_foundation.kernel.resource import Resource
from web_foundation.resources.files.interface import FilesResource
from web_foundation.kernel.dependency import Dependency
from web_foundation.resources.stores.interface import DataStore
from web_foundation.services.service import Service
from web_foundation.utils.helpers import in_obj_subclasses, in_obj_classes


class DependencyContainer:
    app_config: Dependency[GenericConfig] = Dependency(instance_of=BaseSettings)
    data_store: Dependency[DataStore] = Dependency(instance_of=DataStore)
    file_repository: Dependency[FilesResource] = Dependency(instance_of=FilesResource)

    def __init__(self, **kwargs):
        for dep, name in in_obj_classes(self, Dependency):
            dep_init = kwargs.get(name)
            if dep_init:
                dep.from_initialized(dep_init)

    async def init_resources(self, channel: IChannel | None = None):
        for resource, name in in_obj_subclasses(self, Resource):
            await resource.init(self.app_config(), channel)

    async def shutdown_resources(self):
        for resource, name in in_obj_subclasses(self, Resource):
            await getattr(self, name).shutdown()

    def services(self):
        return set(in_obj_subclasses(self, Service))


GenericDependencyContainer = TypeVar("GenericDependencyContainer", bound=DependencyContainer)



from web_foundation.kernel.channel import IMessage


class SettingsChange(IMessage):
    message_type = "change_settings_event"
    name: str
    value: bool

    def __init__(self, name: str, value: bool):
        super().__init__()
        self.name = name
        self.value = value

    def __str__(self):
        return f"{self.__class__.__name__}({self.name})"



from pathlib import Path
from typing import TypeVar

from pydantic import BaseModel, BaseSettings


class DbConfig(BaseModel):
    host: str
    port: str
    database: str
    user: str
    password: str
    db_schema: str
    with_migrations: bool
    migrations_path: Path


class AppConf(BaseSettings):
    app_name: str
    database: DbConfig


GenericConfig = TypeVar("GenericConfig", bound=BaseSettings)



from __future__ import annotations

from functools import partial
from typing import Type, Generic, Callable

import orjson
from sanic import Sanic

from web_foundation.config import AppConf
from web_foundation.extentions.error_handler import ExtendedErrorHandler
from web_foundation.extentions.realtime.rt_broadcaster import RtBroadcaster
from web_foundation.extentions.realtime.rt_connection import RtEventCallback
from web_foundation.extentions.router import ExtRouter
from web_foundation.extentions.sanic_ext_worker import ExtWorker
from web_foundation.kernel.channel import IChannel, IMessage, GenericIMessage
from web_foundation.kernel.container import GenericDependencyContainer
from web_foundation.kernel.dispatcher import IDispatcher, IDispatcherIsolate
from web_foundation.kernel.isolate import Isolate


class App(Generic[GenericDependencyContainer]):
    name: str
    container: GenericDependencyContainer
    sanic: Sanic
    config: AppConf
    dispatcher: IDispatcher
    # listeners: Dict[str, List[AppListener]]
    _rt_listeners = list[tuple[list[Type[IMessage] | str], RtEventCallback, bool]]

    def __init__(self, dependency_container: GenericDependencyContainer,
                 dispatcher: IDispatcher = None, with_addons=False):
        self.name = dependency_container.app_config().app_name
        self.container = dependency_container
        self.sanic = Sanic(self.name, loads=orjson.loads, dumps=orjson.dumps,
                           error_handler=ExtendedErrorHandler())
        self.dispatcher = dispatcher if dispatcher else IDispatcher()
        self._rt_listeners = []
        self.add_isolate("Dispatcher", IDispatcherIsolate, dispatcher=self.dispatcher, add_channel=False, daemon=False)

    def add_isolate(self, name: str, isolate_cls: Type[Isolate], daemon=True, add_channel: bool = True, **kwargs):
        """
        Add process to execute with worker manager
        """
        instance = isolate_cls(name, **kwargs)
        if add_channel:
            self.dispatcher.register_isolate(instance)

        def add_nondaemon(s_app: Sanic):
            nonlocal instance

            def override_manage(ident, func, kwargs, transient=False, daemon=True):
                nonlocal s_app
                container = s_app.manager.transient if transient else s_app.manager.durable
                container.append(
                    ExtWorker(ident, func, kwargs, s_app.manager.context, s_app.manager.worker_state, daemon=daemon))

            s_app.manager.manage = override_manage
            s_app.manager.manage(name, instance.run_forked, {}, daemon=daemon)

        if daemon:
            self.sanic.main_process_ready(
                lambda s_app: s_app.manager.manage(name, instance.run_forked, {}))
        else:
            self.sanic.main_process_ready(lambda s_app: add_nondaemon(s_app))

    def subscribe_worker(self, event_type: list[GenericIMessage | str] | GenericIMessage | str,
                         resolve_callback: RtEventCallback, use_nested_classes: bool = False):
        """
        Subscribe all sanic workers to event[s]
        :param use_nested_classes:
        :param event_type: Name or Class event
        :param resolve_callback: Resolve function
        """
        self._rt_listeners.append((event_type, resolve_callback, use_nested_classes))

    def add_custom_router(self, router: ExtRouter):
        router.ctx.app = self.sanic
        self.sanic.router = router
        self.sanic.router.apply_routes(self.sanic, self.container)

    def _set_sanic_confs(self):
        self.sanic.config.SWAGGER_UI_CONFIGURATION = {
            "docExpansion": 'none'
        }

    def _configure_sanic_main(self):
        async def on_start(s_app: Sanic):
            for i in range(s_app.state.workers):
                channel = IChannel(f"Channel_{i}")
                setattr(s_app.shared_ctx, f"Channel_{i}", channel)
                self.dispatcher.channels.update({f"Channel_{i}": channel})
                # noinspection PyAsyncCall
                self.sanic.add_task(channel.listen_produce(self.dispatcher.on_channel_sent))

        self.sanic.main_process_ready(on_start)

    def _configure_sanic_worker(self):
        async def set_worker_ctx(s_app: Sanic):

            # set channel to ctx
            worker_num = int(s_app.m.name.split('-')[-2])
            s_app.ctx.channel = getattr(s_app.shared_ctx, f"Channel_{worker_num}")
            s_app.ctx.container = self.container
            for service, name in s_app.ctx.container.services():
                service.channel = s_app.ctx.channel
            # add addon loader

            # add broadcaster
            s_app.ctx.rt_broadcaster = RtBroadcaster()
            for event, resolver, use_nested_classes in self._rt_listeners:
                s_app.ctx.channel.add_event_listener(
                    event_type=event,
                    callback=partial(s_app.ctx.rt_broadcaster.broadcast, resolve_callback=resolver),
                    use_nested_classes=use_nested_classes)
            await self.container.init_resources(s_app.ctx.channel)

        self.sanic.before_server_start(set_worker_ctx)

        async def listen_channel(s_app: Sanic):
            # noinspection PyAsyncCall
            s_app.add_task(s_app.ctx.channel.listen_consume())

        self.sanic.after_server_start(listen_channel)

        async def on_stop(s_app: Sanic):
            await self.container.shutdown_resources()

        self.sanic.before_server_stop(on_stop)

    # def _add_listener(self, scope: str, listener: AppListener):
    #     if not self.listeners.get(scope):
    #         self.listeners[scope] = []
    #     self.listeners[scope].append(listener)
    #
    # def before_resources_init(self, listener: AppListener):
    #     self._add_listener("before_resource", listener)
    #
    # def before_app_start(self):
    #     pass

    def run(self, *args, **kwargs):
        self._set_sanic_confs()
        self._configure_sanic_main()
        self._configure_sanic_worker()
        self.sanic.run(*args, **kwargs)


AppListener = Callable[[App], None]



from __future__ import annotations
from dataclasses import dataclass, field

import asyncio
from asyncio import Task
from typing import Generic, Any, Callable, Coroutine

from loguru import logger

from web_foundation import settings
from web_foundation.extentions.request_handler import InputContext
from web_foundation.kernel.channel import GenericIMessage


@dataclass
class RtMessage:
    event_id: str | None = field(default=None)

    def _prepare(self, *args, **kwargs) -> str | bytes:
        raise NotImplementedError

    @property
    def to_sent(self) -> str | bytes:
        return self._prepare()

    @classmethod
    def ping_message(cls):
        raise NotImplementedError


class WriteableObj:
    obj: Any

    def __init__(self, obj: Any):
        self.obj = obj

    async def write(self, message: Any) -> None:
        pass


class RtConnection(Generic[GenericIMessage]):
    ping_timeout: float
    listen_timeout: float
    writeable: WriteableObj
    input_ctx: InputContext
    _last_event_id: int
    _ping_task: Task
    ping_enable: bool
    resolve_callback: RtEventCallback
    _on_disconnect: Callable[[], Coroutine]

    def __init__(self, input_ctx: InputContext, writeable: WriteableObj,
                 resolve_callback: RtEventCallback,
                 ping_enable: bool = False,
                 ping_timeout=None,
                 listen_timeout=None):
        if not resolve_callback:
            raise AttributeError("Resolve Callback can't bee none")
        self.writeable = writeable
        self.resolve_callback = resolve_callback
        self.input_ctx = input_ctx
        self.ping_enable = ping_enable
        self.ping_timeout = ping_timeout if ping_timeout else 5
        self.listen_timeout = listen_timeout if listen_timeout else 0.1
        self._last_event_id = 0

    def _ping_msg(self) -> str | bytes:
        raise NotImplementedError

    async def _ping(self):
        while True:
            if self.ping_enable:
                await self.writeable.write(self._ping_msg())
                # if settings.DEBUG:
                #     real_ip = self.input_ctx.request.headers.get("x-real-ip")
                #     logger.debug(f"Sent ping to {real_ip if real_ip else self.input_ctx.request.ip}")
            await asyncio.sleep(self.ping_timeout)

    async def freeze_request(self, on_disconnect: Callable[[], Coroutine]):
        self._on_disconnect = on_disconnect
        await self._ping()

    async def send_after_call(self, event: GenericIMessage):
        msg = await self.resolve_callback(self, event)
        if msg:
            self._last_event_id += 1
            msg.event_id = self._last_event_id
            await self.writeable.write(msg.to_sent)

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self._on_disconnect()
        if settings.DEBUG:
            real_ip = self.input_ctx.request.headers.get("x-real-ip")
            logger.debug(
                f"RtConnection({self.__class__.__name__}) on {real_ip if real_ip else self.input_ctx.request.ip} disconnected, ping task canceled")

    def _close_condition(self) -> bool:
        raise NotImplementedError

    @classmethod
    async def accept_connection(cls, input_ctx: InputContext,
                                resolve_callback: RtEventCallback = None,
                                ping_enable: bool = True,
                                ping_timeout=None,
                                listen_timeout=None) -> RtConnection:
        raise NotImplementedError

    @classmethod
    async def _construct_writable(cls, input_ctx: InputContext, *args, **kwargs) -> WriteableObj:
        raise NotImplementedError

    @property
    def last_event_num(self):
        return self._last_event_id


RtEventCallback = Callable[[RtConnection, GenericIMessage], Coroutine[Any, Any, RtMessage | None]]



import asyncio
from typing import Callable, Any

from web_foundation.kernel.channel import GenericIMessage
from web_foundation.extentions.realtime.rt_connection import RtEventCallback, RtConnection


class RtBroadcaster:
    """
    Broadcast messages to sse connections of one worker
    """
    rt_connections: list

    def __init__(self):
        self.rt_connections = []

    async def broadcast(self, message: GenericIMessage, resolve_callback: RtEventCallback = None):
        await asyncio.gather(*[
            conn.send_after_call(message)
            for conn in self.rt_connections if
            conn.resolve_callback.__name__ == resolve_callback.__name__
        ])

    async def accept_rt_connection(self, conn: RtConnection,
                                   on_disconnect_callback: Callable[[RtConnection], Any] | None = None):
        self.rt_connections.append(conn)

        async def _on_rt_close():
            nonlocal self
            nonlocal conn
            nonlocal on_disconnect_callback
            self.rt_connections.remove(conn)
            if on_disconnect_callback:
                await on_disconnect_callback(self, conn)

        await conn.freeze_request(_on_rt_close)



import asyncio
from abc import ABCMeta
from typing import Dict

from web_foundation.kernel.channel import IChannel


class Isolate(metaclass=ABCMeta):
    name: str
    channel: IChannel
    init_kwargs: Dict

    def __init__(self, name: str, **kwargs):
        self.name = name
        self.channel = IChannel(f"Channel_{name}")
        self.init_kwargs = kwargs

    async def work(self, **kwargs) -> None:
        raise NotImplementedError

    def run_forked(self) -> None:
        async def call():
            nonlocal self
            asyncio.create_task(self.channel.listen_consume())
            await self.work(**self.init_kwargs)

        try:
            asyncio.run(call())
        except KeyboardInterrupt:
            print("done")
            self.close()

    def close(self):
        pass



import asyncio
from asyncio import Future
from typing import Dict, Any, List

import loguru

from web_foundation import settings
from web_foundation.events.metrics import NewMetricEvent, MetricRequest, MetricResponseEvent
from web_foundation.events.settings import SettingsChange
from web_foundation.kernel.channel import IChannel, EventListener, GenericIMessage
from web_foundation.kernel.isolate import Isolate
from web_foundation.kernel.scheduller import TaskScheduler, TaskIMessage
from web_foundation.metrics.basemetric import BaseMetric, CounterBaseMetric


class IDispatcher:
    channels: Dict[str, IChannel]
    events_listeners: Dict[str, EventListener]
    scheduler: TaskScheduler
    with_scheduler: bool

    collected_metrics: Dict[str, Dict[str, BaseMetric]]
    with_metrics: bool

    _msg_global_index: int

    def __init__(self, with_scheduler: bool = False, with_metrics: bool = False):
        self.with_scheduler = with_scheduler
        self.with_metrics = with_metrics
        self.channels = {}
        self.events_listeners = {}
        self._msg_global_index = 0
        if self.with_scheduler:
            self.scheduler = TaskScheduler(self.broadcast)
            self.add_event_listener(TaskIMessage.message_type, self.scheduler.add_task)
        if with_metrics:
            self.collected_metrics = {}
            self.add_event_listener(NewMetricEvent.message_type, self.on_new_metric)
            self.add_event_listener(MetricRequest.message_type, self.on_metric_request)

        async def on_settings_change(event: SettingsChange):
            setattr(settings, event.name, event.value)

        self.add_event_listener(SettingsChange.message_type, on_settings_change)

    def register_isolate(self, isolate: Isolate):
        self.channels.update({isolate.name: isolate.channel})

    def add_event_listener(self, event_type: str, callback: EventListener):
        self.events_listeners[event_type] = callback

    async def track_event(self, msg: GenericIMessage):
        listener = self.events_listeners.get(msg.message_type)
        if listener:
            await listener(msg)
        if self.with_metrics:
            if settings.EVENTS_METRIC_ENABLE:
                event_counter = CounterBaseMetric("events_counter", self._msg_global_index)
                self.collected_metrics.update({"events_counter": {"dispatcher_events_counter": event_counter}})

                named_event_counter = self.collected_metrics.get("named_events_counter")
                if not named_event_counter:
                    mtr = CounterBaseMetric("named_events_counter", value=1)
                    mtr.add_label(event_name=msg.message_type)
                    named_event_counter = {msg.message_type: mtr}
                    self.collected_metrics["named_events_counter"] = named_event_counter
                elif named_event_counter.get(msg.message_type):
                    named_event_counter[msg.message_type].inc()
                else:
                    mtr = CounterBaseMetric("named_events_counter", value=1)
                    mtr.add_label(event_name=msg.message_type)
                    named_event_counter[msg.message_type] = mtr
                if settings.DEBUG:
                    loguru.logger.debug(f'MetricsDispatcher - track event {msg}')

    async def on_channel_sent(self, msg: GenericIMessage):
        self._msg_global_index += 1
        msg.index = self._msg_global_index
        asyncio.create_task(self.track_event(msg))
        if msg.destination != "__dispatcher__":
            asyncio.create_task(self.broadcast(msg))

    async def on_metric_request(self, event: MetricRequest):
        exporter = event.exporter
        sender_channel = self.channels.get(event.sender)
        if sender_channel:
            if settings.DEBUG:
                loguru.logger.debug(f"Send METRICS to {sender_channel}")
            if not self.collected_metrics:
                await sender_channel.sent_to_consume(exporter.empty())
            rp = []
            for i in self.collected_metrics.values():
                if isinstance(i, dict):
                    for k in i.values():
                        rp.append(k)
                else:
                    rp.append(i)
            resp = MetricResponseEvent(exporter.export(rp))
            resp.inner_index = event.inner_index
            await sender_channel.sent_to_consume(resp)

    async def on_new_metric(self, event: NewMetricEvent):
        metr = self.collected_metrics.get(event.metric.name)
        if not metr:
            self.collected_metrics.update({event.metric.name: {event.metric.labels_concat: event.metric}})
        else:
            if metr.get(event.metric.labels_concat):
                # loguru.logger.warning(f"get {event.metric}")
                metr[event.metric.labels_concat] += event.metric
            else:
                # loguru.logger.warning(f"set {event.metric}")
                metr[event.metric.labels_concat] = event.metric

    async def broadcast(self, msg: GenericIMessage):
        for worker_name, ch in self.channels.items():
            if msg.sender == worker_name:
                continue
            if msg.destination in ["__all__", worker_name]:
                if settings.DEBUG:
                    loguru.logger.debug(f"Sent {msg} to {ch.worker_name}")
                ch.consume_pipe.put(msg)

    def perform(self, **kwargs) -> List[Future[Any]]:
        tasks = []
        for channel in self.channels.values():
            tasks.append(asyncio.ensure_future(
                channel.listen_produce(
                    self.on_channel_sent
                )))
        return tasks


class IDispatcherIsolate(Isolate):
    dispatcher: IDispatcher

    def __init__(self, name: str, dispatcher: IDispatcher, **kwargs):
        self.dispatcher = dispatcher
        super().__init__(name, **kwargs)

    async def work(self, **kwargs) -> None:
        if self.dispatcher.with_scheduler:
            self.dispatcher.scheduler.perform()
        await asyncio.wait(self.dispatcher.perform())



import inspect
import re
from copy import deepcopy
from typing import Dict, Callable, Type

import sanic_ext
from pydantic import BaseModel as PdModel
from sanic_ext.extensions.openapi.builders import SpecificationBuilder, OperationStore


def proccess_definitions(schema: Dict) -> Dict:
    def replace_definition_ref(properties: Dict, definition_name: str):
        for prop_name, value in properties.items():
            if isinstance(value, dict):
                replace_definition_ref(value, definition_name)
            elif isinstance(value, list):
                for s_value in value:
                    if isinstance(s_value, dict):
                        replace_definition_ref(s_value, definition_name)
            if prop_name == "$ref" and f".{definition_name}." in value:
                properties[prop_name] = f"#/components/schemas/{definition_name}"

    definitions = schema.pop("definitions")
    _definitions = {}
    for defin, _schema in definitions.copy().items():
        if "." in defin:
            defin_name = defin.split(".")[-2]
            replace_definition_ref(schema['properties'], defin_name)
            _definitions[defin_name] = _schema
        else:
            _definitions[defin] = _schema
    for defin in _definitions:
        for _schema in _definitions.values():
            if not _schema.get("properties"):
                continue
            replace_definition_ref(_schema['properties'], defin)

    return _definitions


def get_model_schema(model: Type[PdModel]):
    """ Create schema from pydantic model and add all references to OpenAPI schemas """
    schema = model.schema(ref_template="#/components/schemas/{model}")
    components = {}
    if "definitions" in schema:
        definitions = proccess_definitions(schema)
        components.update(definitions)
    components.update({model.__name__: schema})
    spec = SpecificationBuilder()
    for component_name, component in components.items():
        if component_name not in spec._components["schemas"]:
            spec._components["schemas"].update({component_name: component})
    return schema


def add_openapi_spec(uri: str, method_name: str, func: Callable, handler: Callable,
                     in_dto: Type[PdModel] | None, out_dto: Type[PdModel] | None):
    # --- copy existed openapi params ---#
    if OperationStore().get(func):
        OperationStore()[handler] = deepcopy(OperationStore().get(func))
    # --- add query args ---#
    func_text = inspect.getsource(func)
    func_query_args = set(re.findall(r"request.args.get\(\"(\w*)\"\)", func_text))
    if func_query_args:
        for func_arg in func_query_args:
            handler = sanic_ext.openapi.parameter(func_arg)(handler)
    # --- add request body ---#
    if in_dto:
        if issubclass(in_dto, PdModel):
            schema = get_model_schema(in_dto)
            if not hasattr(OperationStore()[handler], "requestBody"):
                OperationStore()[handler].requestBody = {"content": {
                    "environment/json": {'schema': schema},
                    "multipart/form-data": {'schema': schema}
                }}
            elif not OperationStore()[handler].requestBody.get("content"):
                OperationStore()[handler].requestBody['content'] = {
                    "environment/json": {'schema': schema},
                    "multipart/form-data": {'schema': schema}
                }
            if not OperationStore()[handler].requestBody['content'].get("environment/json"):
                OperationStore()[handler].requestBody['content'].update({"environment/json": {'schema': schema}})
            elif not OperationStore()[handler].requestBody['content']["environment/json"].get("schema"):
                OperationStore()[handler].requestBody['content']["environment/json"].update({'schema': schema})
            else:
                for k, v in schema.items():
                    if isinstance(v, dict):
                        OperationStore()[handler].requestBody['content']["environment/json"]['schema'][k].update(v)
                    else:
                        OperationStore()[handler].requestBody['content']["environment/json"]['schema'][k] = v
            if not OperationStore()[handler].requestBody['content'].get("multipart/form-data"):
                OperationStore()[handler].requestBody['content'].update({"multipart/form-data": {'schema': schema}})
            elif not OperationStore()[handler].requestBody['content']["multipart/form-data"].get("schema"):
                OperationStore()[handler].requestBody['content']["multipart/form-data"].update({'schema': schema})
            else:
                for k, v in schema.items():
                    if isinstance(v, dict):
                        OperationStore()[handler].requestBody['content']["multipart/form-data"]['schema'][k].update(v)
                    else:
                        OperationStore()[handler].requestBody['content']["multipart/form-data"]['schema'][k] = v
    # --- add response ---#
    if out_dto:
        if issubclass(out_dto, PdModel):
            schema = get_model_schema(out_dto)
            existed_resp = OperationStore()[handler].responses.get("200")
            if not existed_resp:
                OperationStore()[handler].responses["200"] = {"content": {"environment/json": {'schema': schema}},
                                                              "description": "OK"}
            else:
                if not existed_resp.get('content'):
                    existed_resp.update({"content": {"environment/json": {'schema': schema}},
                                         "description": "OK"})
                elif not existed_resp['content'].get("environment/json"):
                    existed_resp['content'].update({"environment/json": {'schema': schema}})
                elif not existed_resp['content']["environment/json"].get("schema"):
                    existed_resp['content']["environment/json"].update({'schema': schema})
                else:
                    for k, v in schema.items():
                        if isinstance(v, dict):
                            existed_resp['content']["environment/json"]['schema'][k].update(v)
                        else:
                            existed_resp['content']["environment/json"]['schema'][k] = v
    # --- add tag ---#
    handler = sanic_ext.openapi.tag(uri.split("/")[1].capitalize())(handler)  # first path word

    # --- decription --- #
    # f = sanic_ext.openapi.description("hujkiikasdf;awerfl;jkasdfklj;asdfjkl;")(handler)
    # f = sanic_ext.openapi.summary("l;kasdjllww")(handler)

    # --- set operation id --- #
    handler = sanic_ext.openapi.operation(f"{method_name}~{uri}")(handler)
    return handler



from abc import ABC
from dataclasses import dataclass
from functools import wraps
from typing import Any, Callable, Type, Coroutine, TypeVar, Generic
from typing import Dict

from pydantic import BaseModel as PdModel
from sanic import Request, json, HTTPResponse

from web_foundation.kernel.container import GenericDependencyContainer
from web_foundation.extentions.addon_loader import AddonsLoader
from web_foundation.utils.validation import validate_dto

DtoType = TypeVar("DtoType", bound=PdModel)


@dataclass
class ProtectIdentity:
    pass


ProtectIdentityType = TypeVar("ProtectIdentityType", bound=ProtectIdentity)


@dataclass
class InputContext(Generic[DtoType, ProtectIdentityType]):
    request: Request
    dto: DtoType | None
    identity: ProtectIdentity | None
    r_kwargs: Dict


Protector = Callable[[Request, GenericDependencyContainer], Coroutine[Any, Any, ProtectIdentityType | None]]
DtoValidator = Callable[[Type[DtoType], Request], PdModel]
HandlerType = Callable[[InputContext, GenericDependencyContainer], Coroutine[Any, Any, HTTPResponse]]


class ExtRequestHandler(ABC):
    protector: Protector | None
    in_struct: Type[PdModel] | None
    addons_loader: AddonsLoader | None
    validation_fnc: DtoValidator | None
    container: GenericDependencyContainer | None
    response_fabric: Callable[[Any], HTTPResponse] | None

    def __init__(self,
                 container: GenericDependencyContainer = None,
                 protector: Protector = None,
                 in_struct: Type[PdModel] | None = None,
                 addon_manager: AddonsLoader = None,
                 validation_fnc: DtoValidator = validate_dto,
                 response_fabric: Callable[[Any], HTTPResponse] = json):
        self.protector = protector
        self.in_struct = in_struct
        self.addons_loader = addon_manager
        self.validation_fnc = validation_fnc
        self.response_fabric = response_fabric
        self.container = container

    def __call__(self, target: HandlerType):
        raise NotImplementedError


GenericRequestHandler = TypeVar('GenericRequestHandler', bound=ExtRequestHandler)


class ChainRequestHandler(ExtRequestHandler):

    def __call__(self, target: HandlerType):
        @wraps(target)
        async def f(*args, **kwargs):
            req: Request = args[0]
            prot_identity = await self.protector(req, self.container) if self.protector else None
            if self.in_struct:
                validated = self.validation_fnc(self.in_struct, req)
            else:
                validated = None
            incoming = InputContext(req, validated, prot_identity, kwargs)
            if self.addons_loader:
                plugins = await self.addons_loader.find_middleware_by_target(target.__name__)
                for plugin in await self.addons_loader.find_middleware_by_target(target.__name__):
                    await plugin.exec_before(context=incoming, container=self.container)
                if not plugins:
                    ret_val = await target(incoming, self.container)
                else:
                    overrides = [pl for pl in plugins if pl.override]
                    if overrides:
                        ret_val = await overrides[0].exec_override(context=incoming, container=self.container)
                    else:
                        ret_val = await target(incoming, self.container)
                    for plugin in plugins:
                        await plugin.exec_after(context=incoming, container=self.container, target_result=ret_val)
            else:
                ret_val = await target(incoming, self.container)
            if isinstance(ret_val, HTTPResponse):
                return ret_val
            return self.response_fabric(ret_val)

        return f


"""
def chain(protector: Protector = None,
          in_struct: Type[PdModel] | None = None,
          plugin_manager: AddonsLoader = None,
          validation_fnc: DtoValidator = validate_dto,
          response_fabric: Callable[[TypeJSON], HTTPResponse] = json,
          container: Container = None
          ):
    def called_method(target: HandlerType):
        @wraps(target)
        async def f(*args, **kwargs):
            req: Request = args[0]
            prot_identity = await protector(req) if protector else None
            if in_struct:
                validated = validation_fnc(in_struct, req)
            else:
                validated = None
            incoming = InputContext(req, validated, prot_identity, kwargs)
            if plugin_manager:
                plugins = await plugin_manager.find_middleware_by_target(target.__name__)
                for plugin in await plugin_manager.find_middleware_by_target(target.__name__):
                    await plugin.exec_before(context=incoming, container=container)
                if not plugins:
                    ret_val = await target(incoming, container)
                else:
                    overrides = [pl for pl in plugins if pl.override]
                    if overrides:
                        ret_val = await overrides[0].exec_override(context=incoming, container=container)
                    else:
                        ret_val = await target(incoming, container)
                    for plugin in plugins:
                        await plugin.exec_after(context=incoming, container=container, target_result=ret_val)
            else:
                ret_val = await target(incoming, container)
            return response_fabric(ret_val)

        return f

    return called_method
"""



from typing import Any

from web_foundation.metrics.basemetric import BaseMetric
from web_foundation.metrics.exporter import MetricExporter
from web_foundation.kernel.channel import IMessage


class NewMetricEvent(IMessage):
    message_type = "new_metric"
    destination = "__dispatcher__"
    metric: BaseMetric

    def __init__(self, metric: BaseMetric):
        super().__init__()
        self.metric = metric

    def __str__(self):
        return f"NewMetricEvent({self.metric.__str__()})"


class MetricRequest(IMessage):
    message_type = "metrics_request"
    destination = "__dispatcher__"
    exporter: MetricExporter

    def __init__(self, exporter: MetricExporter):
        super().__init__()
        self.exporter = exporter


class MetricResponseEvent(IMessage):
    message_type = "metrics_response"
    metrics_data: Any

    def __init__(self, metrics_data: Any):
        super().__init__()
        self.metrics_data = metrics_data

    def __str__(self):
        return f"MetricResponse({type(self.metrics_data)})"



from __future__ import annotations

import asyncio
from multiprocessing import Queue
from typing import Coroutine, Callable, Dict, Type, List, Any
from typing import TypeVar

from aioprocessing import AioQueue
from loguru import logger

from web_foundation import settings


class IMessage:
    message_type: str = "__all__"
    index: int
    inner_index: int
    sender: str
    destination: str = "__all__"
    exec_inner: bool = False

    def __init__(self):
        self.index = 0
        self.inner_index = 0
        self.sender = "None"

    def __str__(self):
        return f"{self.__class__.__name__}({self.message_type},{self.index=},{self.inner_index=}, {self.sender=})"


GenericIMessage = TypeVar("GenericIMessage", bound=IMessage, contravariant=True)

EventListener = Callable[[GenericIMessage], Coroutine]


class IMessageWaiter:
    event = asyncio.Event()
    result: Any = None
    call = None

    def __eq__(self, other):
        self.wait_for_index = other.wait_for_index


class IChannel:
    read_timeout = 0.01
    worker_name: str
    consume_pipe: AioQueue
    produce_pipe: Queue

    _listeners: Dict[str, List[EventListener]]
    _response_waiters: Dict[int, IMessageWaiter]

    _before_event_middlewares: Dict[str, EventListener]
    _after_event_middlewares: Dict[str, EventListener]
    _inner_index: int

    def __init__(self, isolate_name: str):
        self._inner_index = 0
        self.worker_name = isolate_name
        # self.debug = debug
        self._listeners = {}
        self._response_waiters = {}
        self.consume_pipe = Queue()
        self.produce_pipe = Queue()
        self._after_event_middlewares = {}
        self._before_event_middlewares = {}

    def __str__(self):
        return f"IChannel(name={self.worker_name})"

    async def produce(self, msg: IMessage):
        self._inner_index += 1
        msg.inner_index = self._inner_index
        msg.sender = self.worker_name
        before_middleware = self._before_event_middlewares.get(msg.message_type)
        if before_middleware:
            await before_middleware(msg)
        if msg.exec_inner:
            listeners = self._listeners.get(msg.message_type)
            for listener in listeners:
                asyncio.create_task(listener(msg))
        self.produce_pipe.put(msg)

    async def sent_to_consume(self, msg: IMessage):
        self.consume_pipe.put(msg)

    async def listen_produce(self, callback: Callable[[GenericIMessage], Coroutine]):
        while True:
            while self.produce_pipe.empty():
                await asyncio.sleep(0.1)
            r: GenericIMessage = self.produce_pipe.get()
            if settings.DEBUG:
                logger.info(
                    f"Channel {self.worker_name} send message: {r.message_type}")
            await callback(r)

    async def listen_consume(self):
        while True:
            while self.consume_pipe.empty():
                await asyncio.sleep(0.1)
            r: IMessage = self.consume_pipe.get()
            if settings.DEBUG:
                logger.info(
                    f"Channel {self.worker_name} receive message: {r.message_type}")
            after_middleware = self._after_event_middlewares.get(r.message_type)
            if after_middleware:
                asyncio.create_task(after_middleware(r))
            response_waiter = self._response_waiters.get(r.inner_index)
            if response_waiter:
                response_waiter.result = r
            callbacks = self._listeners.get(r.message_type)
            if not callbacks:
                continue
            for callback in callbacks:
                asyncio.create_task(callback(r))

    def add_event_middleware(self, event_type: str, callback: EventListener, assign_to: str = "after"):
        """
        :param callback: callback to call
        :param event_type: "event_message_type"
        :param assign_to: "before, after"
        :return: None
        """
        match assign_to:
            case 'before':
                self._before_event_middlewares.update({event_type: callback})
            case "after":
                self._after_event_middlewares.update({event_type: callback})

    async def produce_for_response(self, msg: GenericIMessage) -> Any:
        task = IMessageWaiter()
        await self.produce(msg)
        self._response_waiters.update({msg.inner_index: task})
        while not task.result:
            await asyncio.sleep(0.01)
        self._response_waiters.pop(msg.inner_index)
        return task.result

    def _add_event_listener(self, event_type: Type[IMessage] | str, callback: EventListener,
                            use_nested_classes: bool = False):
        if use_nested_classes and isinstance(event_type, str):
            raise AttributeError("Can't add_event_listener with use_nested_classes")

        def _add(event_name: str):
            nonlocal self
            nonlocal callback
            if event_name not in self._listeners:
                self._listeners[event_name] = []
            self._listeners[event_name].append(callback)

        if isinstance(event_type, str):
            _add(event_type)
        else:
            def _raise(cls_type):
                if not hasattr(cls_type, "message_type"):
                    raise AttributeError("Can't register listener cause message_type not found")

            if use_nested_classes:
                for cls in event_type.__subclasses__():
                    _raise(cls)
                    _add(cls.message_type)
            else:
                _raise(event_type)
                _add(event_type.message_type)

    def add_event_listener(self, event_type: List[Type[IMessage] | str] | Type[IMessage] | str,
                           callback: EventListener, use_nested_classes: bool = False):
        if isinstance(event_type, list):
            for i in event_type:
                self._add_event_listener(i, callback, use_nested_classes)
        else:
            self._add_event_listener(event_type, callback, use_nested_classes)

    def remove_event_listeners(self, event_type: Type[IMessage] | str):
        if isinstance(event_type, str):
            self._listeners.pop(event_type)
        else:
            self._listeners.pop(event_type.message_type)

    @property
    def listeners(self):
        return self._listeners



import os
from enum import Enum
from pathlib import Path
from typing import Type, TypeVar, Generic, Any, Generator
from aiofiles.threadpool import open
from aiofiles.base import AiofilesContextManager

from web_foundation.kernel.channel import IChannel
from web_foundation.resources.files.interface import FilesResource

class OsFilesResource(FilesResource):

    def __init__(self, root, *args, **kwargs):
        self.root = root
        super().__init__(*args, **kwargs)

    def exists(self, path: Path) -> bool:
        return self.root.joinpath(path).exists()

    def get_full_path(self, path: Path) -> Path:
        return self.root.joinpath(path)

    async def open(self, path: Path, *args, **kwargs) -> AiofilesContextManager:
        return open(self.root.joinpath(path), *args, **kwargs)

    async def list(self, path: Path, *args, **kwargs) -> Generator[Path, None, None]:
        target = self.root.joinpath(path)
        return target.iterdir()

    async def shutdown(self):
        pass

    async def init(self, channel: IChannel, *args, **kwargs):
        await super(OsFilesResource, self).init(channel, *args, **kwargs)
        if not self.root.exists():
            raise FileNotFoundError(str(self.root))



from __future__ import annotations

import asyncio
import datetime
import os
import uuid
from dataclasses import dataclass, field
from functools import partial
from time import time_ns
from typing import Dict, Callable, Coroutine, Any, TypeVar, Generic

import loguru
from apscheduler.events import EVENT_JOB_SUBMITTED, JobSubmissionEvent, JobExecutionEvent, \
    EVENT_JOB_ERROR, EVENT_JOB_EXECUTED
from apscheduler.executors.base import BaseExecutor
from apscheduler.executors.pool import ProcessPoolExecutor
from apscheduler.job import Job
from apscheduler.jobstores.memory import MemoryJobStore
from apscheduler.schedulers.background import BackgroundScheduler
from apscheduler.triggers.base import BaseTrigger
from apscheduler.triggers.interval import IntervalTrigger
from pytz import utc

from web_foundation.kernel.channel import IMessage
from web_foundation.utils.helpers import get_callable_from_fnc

BackgroundTask = TypeVar("BackgroundTask", Callable[..., Coroutine[Any, Any, Any]], partial)


@dataclass
class Task:
    id: str
    status: str
    name: str
    scheduled_trigger: BaseTrigger
    scheduled_job: Job | None = field(default=None)
    on_error_callback: TaskErrorCallback = field(default=None)
    next_call_time: datetime.datetime | None = field(default=None)
    execution_time: float = field(default=0)
    done_time: float = field(default=0)
    call_time: float = field(default=0)
    call_counter: int = field(default=0)
    create_time: datetime.datetime | None = field(default_factory=datetime.datetime.now)
    error: BaseException | None = field(default=None)
    return_event: bool = False


TaskErrorCallback = Callable[[JobExecutionEvent, Task], Coroutine[Any, Any, None]]


class TaskIMessage(IMessage):
    message_type = "background_task"
    task: BackgroundTask
    args: Any | None
    kwargs: Any | None
    trigger: BaseTrigger
    add_job_kw: Dict
    on_error_callback: TaskErrorCallback
    return_event: bool
    destination = "__dispatcher__"

    def __init__(self, task: BackgroundTask,
                 trigger=None,
                 args: Any = None,
                 kwargs: Any | None = None,
                 add_job_kw: Dict = None,
                 on_error_callback: TaskErrorCallback = None,
                 return_event: bool = False):
        super().__init__()
        self.task = task
        self.args = args
        self.kwargs = kwargs
        self.add_job_kw = add_job_kw if add_job_kw else {}
        self.trigger = trigger
        self.on_error_callback = on_error_callback
        self.return_event = return_event


ProduceCall = Callable[[TaskIMessage], Coroutine[Any, Any, None]]


class TaskScheduler:
    _tasks: Dict[str, Task]
    _scheduler: BackgroundScheduler
    produce_method: ProduceCall

    def __init__(self, produce_method: ProduceCall, executors: Dict[str, BaseExecutor] = None, job_defaults=None,
                 **kwargs):
        self._tasks = {}
        self.produce_method = produce_method
        executor = {
            'default': ProcessPoolExecutor(os.cpu_count())
        }
        job_default = {
            "misfire_grace_time": 1,
            'coalesce': False,
            'max_instances': 10
        }
        self._scheduler = BackgroundScheduler(jobstores={"default": MemoryJobStore()},
                                              executors=executors if executors else executor,
                                              job_defaults=job_defaults if job_defaults else job_default,
                                              timezone=utc)
        self._add_listeners()

    async def add_task(self, message: TaskIMessage):
        message.task = get_callable_from_fnc(message.task)

        if isinstance(message.trigger, IntervalTrigger):
            for task_id, task in self._tasks.items():
                if task.scheduled_job.func == message.task:
                    task.scheduled_job.trigger = message.trigger
                    task.scheduled_trigger = message.trigger
                    return

        arc_task = Task(id=str(uuid.uuid4()),
                        name=message.task.__name__,
                        status="new",
                        scheduled_trigger=message.trigger,
                        on_error_callback=message.on_error_callback,
                        return_event=message.return_event)
        arc_task.scheduled_job = self._scheduler.add_job(message.task,
                                                         name=arc_task.name,
                                                         id=arc_task.id,
                                                         trigger=message.trigger,
                                                         args=message.args,
                                                         kwargs=message.kwargs,
                                                         **message.add_job_kw)
        loguru.logger.warning(arc_task.scheduled_job)
        loguru.logger.warning(os.getpid())
        arc_task.next_call_time = arc_task.scheduled_job.next_run_time
        self._tasks.update({arc_task.id: arc_task})

    def _add_listeners(self):
        def _on_job_submitted(event: JobSubmissionEvent):
            nonlocal self
            task = self._tasks.get(event.job_id)
            task.call_time = time_ns()
            task.call_counter += 1
            task.status = "called"

        def _on_job_exec(event: JobExecutionEvent):
            nonlocal self
            task = self._tasks.get(event.job_id)
            task.done_time = time_ns()
            task.execution_time = task.done_time - task.call_time
            task.status = "done"
            task.next_call_time = task.scheduled_job.next_run_time
            if event.exception:
                task.error = event.exception
                task.status = 'error'
                if task.on_error_callback:
                    asyncio.run(task.on_error_callback(event, task))
            elif task.return_event and event.retval:
                asyncio.run(self.produce_method(event.retval))

        self._scheduler.add_listener(_on_job_submitted, EVENT_JOB_SUBMITTED)
        self._scheduler.add_listener(_on_job_exec, EVENT_JOB_EXECUTED | EVENT_JOB_ERROR)

    def close(self, *args, **kwargs):
        self._scheduler.shutdown()

    def perform(self):
        self._scheduler.start()
